/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

//  This class defines address settings applied to a VM
//  by the GCS every time a VM starts or restores.
type HvSocketAddress struct {
	LocalAddress  string `json:"LocalAddress,omitempty"`
	ParentAddress string `json:"ParentAddress,omitempty"`
}
