/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type IovSettings struct {
	// The weight assigned to this port for I/O virtualization (IOV) offloading.
	// Setting this to 0 disables IOV offloading.
	OffloadWeight *uint32 `json:"OffloadWeight,omitempty"`

	// The number of queue pairs requested for this port for I/O virtualization (IOV) offloading.
	QueuePairsRequested *uint32 `json:"QueuePairsRequested,omitempty"`

	// The interrupt moderation mode for I/O virtualization (IOV) offloading.
	InterruptModeration *InterruptModerationName `json:"InterruptModeration,omitempty"`
}
