/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type ConsoleSize struct {
	Height int32 `json:"Height,omitempty"`

	Width int32 `json:"Width,omitempty"`
}
