/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type DeviceExtension struct {
	DeviceCategory *DeviceCategory           `json:"device_category,omitempty"`
	Namespace      *DeviceExtensionNamespace `json:"namespace,omitempty"`
}
